import 'dart:async';
import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

class ListScaffoldWidget<T extends IdItemInterface> extends StatefulWidget {
  final Widget title;
  final List<T> list;
  final List<T> Function(BuildContext context, bool searchActive) listBuilder;
  final List<Widget> appBarButtons;
  final List<Widget> appBarPopupButtons;
  final bool showAppBarButtonsOnSelection;
  final bool outsourceMenuButtons;
  final ISelectableListItem Function(BuildContext context, State state, T item,
      bool isItemPicker, List<String> searchTexts) itemBuilder;
  final Comparable Function(T) sortProperty;
  final List<Comparable> Function(T) sortProperties;
  final bool reverseSort;
  final Comparable Function(T) groupProperty;
  final Widget Function(T) groupTitle;
  final List<Widget> Function(BuildContext context,
          ListScaffoldWidgetState<T> listScaffoldWidgetState)
      selectionMenuButtonsBuilder;
  final bool titleFullSpace;
  final Future<bool> Function() onWillPop;
  final bool itemsAreSelectable;
  final List<Widget> menuButtons;
  final Widget topWidget;
  final Widget Function(
          BuildContext context, bool isItemPicker, bool searchActive)
      topWidgetBuilder;
  final Widget noItemWidget;
  final Widget Function(
          BuildContext context, bool isItemPicker, bool searchActive)
      noItemWidgetBuilder;
  final bool showBackArrow;
  final Widget selectionTitle;
  final bool showBackArrowOnSelection;
  final IHelpWidget helpWidget;

  final bool isItemPicker;
  final bool multiSelectPicker;
  final List<DocumentReference> selectedItems;
  final String itemCollectionNameForPicker;
  final void Function(List<DocumentReference> refs) saveRefsToItem;

  final ListScaffoldScrollBarConfiguration<T> scrollBar;

  final bool withSearch;
  final bool searchDefaultActive;
  final List<String> Function(T item) searchArguments;

  const ListScaffoldWidget({
    Key key,
    this.title,
    this.itemBuilder,
    this.list,
    this.listBuilder,
    this.sortProperty,
    this.groupProperty,
    this.appBarButtons,
    this.showAppBarButtonsOnSelection = false,
    this.selectionMenuButtonsBuilder,
    List<Widget> selectionMenuButtons,
    this.outsourceMenuButtons = false,
    this.groupTitle,
    this.titleFullSpace = false,
    this.itemsAreSelectable = true,
    this.menuButtons,
    this.topWidget,
    this.showBackArrow = false,
    this.reverseSort = false,
    this.sortProperties,
    this.noItemWidget,
    this.isItemPicker = false,
    this.multiSelectPicker,
    this.selectedItems,
    this.saveRefsToItem,
    this.selectionTitle,
    this.helpWidget,
    this.onWillPop,
    this.scrollBar,
    this.showBackArrowOnSelection = true,
    this.appBarPopupButtons,
    this.withSearch = false,
    this.searchArguments,
    this.noItemWidgetBuilder,
    this.topWidgetBuilder,
    this.itemCollectionNameForPicker,
    this.searchDefaultActive = false,
  })  : assert(list != null || listBuilder != null),
        assert(itemBuilder != null),
        assert(!withSearch || searchArguments != null),
        assert(sortProperty == null || sortProperties == null),
        assert(isItemPicker == false || multiSelectPicker != null),
        assert(isItemPicker == false || selectedItems != null),
        assert(isItemPicker == false || saveRefsToItem != null),
        assert(isItemPicker == false || itemCollectionNameForPicker != null),
        super(key: key);

  @override
  ListScaffoldWidgetState<T> createState() => ListScaffoldWidgetState<T>();
}

class ListScaffoldWidgetState<T extends IdItemInterface>
    extends State<ListScaffoldWidget<T>> {
  bool selectionActive;
  List<T> selectedItems;

  bool searchActive;
  TextEditingController searchController = TextEditingController(text: "");
  FocusNode searchFocusNode = FocusNode();

  Map<T, double> searchResults;
  List<List<T>> groupedItems;

  FocusNode focusNode = FocusNode();

  ScrollController scrollController = ScrollController();

  @override
  void initState() {
    super.initState();

    selectionActive = widget.isItemPicker;
    searchActive = widget.searchDefaultActive;
    if (searchActive)
      Future.delayed(Duration(milliseconds: 100),
          () => FocusScope.of(context).requestFocus(searchFocusNode));
    groupedItems = [];
    selectedItems = <T>[];
    searchController.addListener(() => setState(() {}));

    if (widget.isItemPicker) {
      Future.delayed(Duration.zero, () {
        selectedItems = (widget.list ??
                widget.listBuilder(context, searchActive))
            .where(
                (item) => widget.selectedItems.any((ref) => ref.id == item.id))
            .toList();
        if (selectedItems.length != 0) setState(() {});
      });
    }

    CleanAndroidDesignSettings.connectionChange?.listen((_) => setState(() {}));
  }

  @override
  Widget build(BuildContext context) {
    if (searchActive)
      _searchList();
    else
      _sortList();

    return RawKeyboardListener(
      focusNode: focusNode,
      autofocus: true,
      onKey: (event) {
        if (event.logicalKey == LogicalKeyboardKey.keyS &&
            !event.isControlPressed &&
            !searchActive) {
          setState(() {
            searchActive = true;
            searchController.clear();
            Future.delayed(Duration(milliseconds: 100),
                () => FocusScope.of(context).requestFocus(searchFocusNode));
          });
        }
        if (event.logicalKey == LogicalKeyboardKey.escape &&
            (searchActive || selectionActive))
          setState(() {
            searchActive = false;
            selectionActive = false;
            selectedItems.clear();
          });
      },
      child: DefaultScaffoldWidget(
        title: searchActive ? _buildSearchBar() : _getTitleWidget(),
        appBarButtons:
            ((selectionActive && !widget.showAppBarButtonsOnSelection) ||
                    searchActive)
                ? null
                : (widget.withSearch
                    ? ([]
                      ..addAll(widget.appBarButtons ?? [])
                      ..add(searchAppBarButton()))
                    : widget.appBarButtons),
        appBarPopupButtons:
            (selectionActive && !widget.showAppBarButtonsOnSelection)
                ? null
                : widget.appBarPopupButtons,
        showBackArrow: (widget.showBackArrow ||
                (selectionActive && widget.showBackArrowOnSelection)) &&
            !searchActive,
        outsourceMenuButtons: widget.outsourceMenuButtons,
        titleFullSpace: widget.titleFullSpace || searchActive,
        backgroundFocus: focusNode,
        onWillPop: () {
          if (selectionActive && !widget.isItemPicker) {
            cancelSelection();
            return Future.value(false);
          } else
            return widget.onWillPop?.call() ?? Future.value(true);
        },
        scrollable: false,
        helpWidget: widget.helpWidget,
        child: (groupedItems.length == 0 ||
                (groupedItems.length == 1 && groupedItems[0].length == 0)
            ? _buildNoItemWidget()
            : _getListView()),
        menuButtons: widget.isItemPicker
            ? _buildItemPickerMenuButtons()
            : (selectionActive && widget.selectionMenuButtonsBuilder != null
                ? widget.selectionMenuButtonsBuilder(context, this)
                : widget.menuButtons),
        scrollBar: searchActive ? null : getScrollBarConfiguration(),
      ),
    );
  }

  Widget searchAppBarButton() {
    return IconButton(
      icon: Icon(
        Icons.search,
      ),
      onPressed: () {
        setState(() {
          searchActive = true;
          searchController.clear();
          Future.delayed(Duration(milliseconds: 100),
              () => FocusScope.of(context).requestFocus(searchFocusNode));
        });
      },
      tooltip: "Suchen",
    );
  }

  DefaultScaffoldScrollBarConfiguration<T> getScrollBarConfiguration() {
    if (widget.scrollBar == null) return null;

    var sectionItems = <Widget>[];
    var sectionIndicatorItems =
        widget.scrollBar.groupScrollBarIndicatorWidget != null
            ? <Widget>[]
            : null;
    var sectionPositions = <double>[];
    var nextPos = widget.scrollBar.scrollBarOffset;
    groupedItems.forEach((group) {
      sectionItems.add(widget.scrollBar.groupScrollBarWidget(group.first));
      sectionIndicatorItems
          ?.add(widget.scrollBar.groupScrollBarIndicatorWidget(group.first));
      sectionPositions.add(nextPos);
      nextPos += widget.scrollBar.sectionHeaderHeight +
          group.length * widget.scrollBar.listItemHeight;
    });

    return DefaultScaffoldScrollBarConfiguration<T>(
      groupScrollBarIndicatorWidth:
          widget.scrollBar.groupScrollBarIndicatorWidth,
      sectionItems: sectionItems,
      sectionIndicatorItems: sectionIndicatorItems,
      sectionPositions: sectionPositions,
      scrollController: scrollController,
    );
  }

  List<Widget> _buildItemPickerMenuButtons() {
    var list = <Widget>[
      SimpleScaffoldMenuButtonWidget(
        text: Text("Abbrechen"),
        onTap: () => Navigator.pop(context),
        tooltip: "Auswahl abbrechen",
      ),
    ];

    if (widget.multiSelectPicker)
      list.add(SimpleScaffoldMenuButtonWidget(
        text: Text("Auswählen"),
        onTap: () => _savePickedItemsToItem(),
        tooltip: "Auswahl speichern",
      ));

    return list;
  }

  _savePickedItemsToItem() {
    var refList = selectedItems
        .map((item) => CleanAndroidDesignSettings.firestore
            .doc("${widget.itemCollectionNameForPicker}/${item.id}"))
        .toList();
    widget.saveRefsToItem(refList);

    Navigator.pop(context);
  }

  void _searchList() {
    var searchTexts = searchController.text.toLowerCase().split(" ")
      ..removeWhere((s) => s == "");

    var list = widget.list ?? widget.listBuilder(context, searchActive);

    searchResults = Map<T, double>();
    if (searchTexts.length > 0) {
      searchResults.updateAll((key, value) => 0);
      list.forEach((item) => searchResults.update(
            item,
            (_) => _calcSearchResults(item, searchTexts),
            ifAbsent: () => _calcSearchResults(item, searchTexts),
          ));
    } else {
      searchResults.updateAll((key, value) => 0);
    }

    var results = list.where((item) => (searchResults[item] ?? 0) > 0).toList()
      ..sort((item1, item2) {
        if (searchResults[item2].compareTo(searchResults[item1]) != 0)
          return searchResults[item2].compareTo(searchResults[item1]);

        if (widget.sortProperty != null)
          list.sort((item1, item2) => widget
              .sortProperty(widget.reverseSort ? item2 : item1)
              .compareTo(
                  widget.sortProperty(widget.reverseSort ? item1 : item2)));
        if (widget.sortProperties != null) {
          var p1 = widget.sortProperties(widget.reverseSort ? item2 : item1);
          var p2 = widget.sortProperties(widget.reverseSort ? item1 : item2);
          for (var i = 0; i < p1.length; i++) {
            var result = p1[i].compareTo(p2[i]);
            if (result != 0) return result;
          }
        }
        return 0;
      });
    groupedItems = results.length == 0 ? [] : [results];
  }

  double _calcSearchResults(T item, List<String> searchTexts) {
    if (searchTexts.length == 0 || widget.searchArguments == null) return 1;
    var scores = List.generate(searchTexts.length, (index) => 0.0);
    searchTexts.asMap().forEach((n, searchText) {
      var args = widget
          .searchArguments(item)
          .map((t) => (t ?? "").toLowerCase())
          .toList();
      for (var i = 0; i < args.length; i++) {
        if (args[i].contains(searchText)) {
          if (args[i].startsWith(searchText)) {
            scores[n] = max(scores[n], (args.length - i) * 10 + 1.0);
          } else {
            scores[n] = max(scores[n], (args.length - i) * 10 + 0.5);
          }
        }
      }
    });
    return scores.reduce(min) == 0 ? 0 : scores.reduce(max);
  }

  void _sortList() {
    var list = widget.list ?? widget.listBuilder(context, searchActive);
    if (widget.sortProperty != null)
      list.sort((item1, item2) => widget
          .sortProperty(widget.reverseSort ? item2 : item1)
          .compareTo(widget.sortProperty(widget.reverseSort ? item1 : item2)));
    if (widget.sortProperties != null) {
      list.sort((item1, item2) {
        var p1 = widget.sortProperties(widget.reverseSort ? item2 : item1);
        var p2 = widget.sortProperties(widget.reverseSort ? item1 : item2);
        for (var i = 0; i < p1.length; i++) {
          var result = p1[i].compareTo(p2[i]);
          if (result != 0) return result;
        }
        return 0;
      });
    }

    groupedItems = <List<T>>[];
    if (widget.groupProperty == null) {
      groupedItems.add(list);
    } else if (list.length > 0) {
      var lastList = <T>[];
      groupedItems.add(lastList);
      list.asMap().forEach((i, item) {
        if (i == 0)
          lastList.add(item);
        else if (widget
                .groupProperty(lastList.last)
                .compareTo(widget.groupProperty(item)) ==
            0)
          lastList.add(item);
        else {
          lastList = <T>[];
          lastList.add(item);
          groupedItems.add(lastList);
        }
      });
    }
  }

  Widget _buildNoItemWidget() {
    return Column(
      children: <Widget>[
        widget.topWidget ??
            (widget.topWidgetBuilder != null
                ? widget.topWidgetBuilder(
                    context, widget.isItemPicker, searchActive)
                : null) ??
            Container(),
        DescriptionText.fromChild(widget.noItemWidget ??
            (widget.noItemWidgetBuilder != null
                ? widget.noItemWidgetBuilder(
                    context, widget.isItemPicker, searchActive)
                : (searchActive
                    ? (searchController.text.length == 0
                        ? Text("Bitte einen Suchtext eingeben")
                        : Text("Kein Suchergebnis gefunden"))
                    : null))),
      ],
    );
  }

  Widget _getListView() {
    var topWidget = widget.topWidget ??
        (widget.topWidgetBuilder == null
            ? null
            : widget.topWidgetBuilder(
                context, widget.isItemPicker, searchActive));
    var offset = topWidget == null ? 0 : 1;
    return ListView.builder(
      controller: scrollController,
      itemCount: groupedItems.length + offset,
      itemBuilder: (context, groupIndex) {
        if (topWidget != null && groupIndex == 0) return topWidget;

        return ItemSectionWidget.builder(
          title: (widget.groupProperty == null || searchActive)
              ? null
              : widget.groupTitle == null
                  ? Text(widget
                      .groupProperty(groupedItems[groupIndex - offset].first)
                      .toString())
                  : widget.groupTitle(groupedItems[groupIndex - offset].first),
          itemCount: groupedItems[groupIndex - offset].length,
          itemBuilder: (context, itemIndex) {
            var item = groupedItems[groupIndex - offset][itemIndex];
            return widget
                .itemBuilder(
                  context,
                  this,
                  item,
                  widget.isItemPicker,
                  searchActive
                      ? (searchController.text.toLowerCase().split(" ")
                        ..removeWhere((s) => s == ""))
                      : null,
                )
                .copyWithSelection(
                    isSelected: selectedItems.contains(item),
                    onTapList: () {
                      if (selectionActive)
                        setState(() {
                          if (widget.isItemPicker && !widget.multiSelectPicker)
                            selectedItems.clear();
                          if (!selectedItems.remove(item))
                            selectedItems.add(item);
                          if (!widget.multiSelectPicker)
                            _savePickedItemsToItem();
                        });
                      return !selectionActive;
                    },
                    onLongTapList: widget.isItemPicker
                        ? null
                        : () {
                            if (widget.itemsAreSelectable)
                              setState(() {
                                selectionActive = true;
                                if (!selectedItems.remove(item))
                                  selectedItems.add(item);
                              });
                          });
          },
        );
      },
    );
  }

  Widget _getTitleWidget() {
    return selectionActive
        ? widget.selectionTitle ??
            Text(
              selectedItems.length.toString() + " ausgewählt",
            )
        : widget.title;
  }

  Widget _buildSearchBar() {
    return Container(
      width: MediaQuery.of(context).size.width,
      height: 42,
      decoration: BoxDecoration(
        borderRadius: BorderRadius.circular(21),
        color: Colors.grey[300],
      ),
      child: Row(
        crossAxisAlignment: CrossAxisAlignment.center,
        children: <Widget>[
          IconButton(
            icon: Icon(
              Icons.keyboard_arrow_left,
              color: CleanAndroidDesignDefaultColors.black,
              size: 30,
            ),
            onPressed: () {
              setState(() {
                searchActive = false;
              });
            },
          ),
          Expanded(
            child: TextField(
              controller: searchController,
              cursorColor: CleanAndroidDesignDefaultColors.accent,
              decoration: InputDecoration(
                hintText: "Suche",
                border: InputBorder.none,
              ),
              textCapitalization: TextCapitalization.sentences,
              textInputAction: TextInputAction.search,
              style: TextStyles.textStyleSize(18)
                  .copyWith(color: CleanAndroidDesignDefaultColors.black),
              focusNode: searchFocusNode,
            ),
          ),
          searchController.text != ""
              ? IconButton(
                  icon: Icon(
                    Icons.close,
                    color: CleanAndroidDesignDefaultColors.black,
                    size: 25,
                  ),
                  onPressed: () => setState(() => searchController.text = ""),
                )
              : Container(),
        ],
      ),
    );
  }

  void cancelSelection() {
    setState(() {
      selectionActive = false;
      selectedItems.clear();
    });
  }
}
